/* The modifiers XEmacs knows about; these appear in key and button events. */

#define MOD_CONTROL	(1<<0)
#define MOD_META	(1<<1)
#define MOD_SUPER	(1<<2)
#define MOD_HYPER	(1<<3)
#define MOD_ALT		(1<<4)
#define MOD_SHIFT	(1<<5)  /* not used for dual-case characters */
